#include "lua_volksdata.h"
#include "stackdump.h"


VOLK_Term **allocate_term (lua_State *L)
{
    VOLK_Term **tp = lua_newuserdatauv (L, sizeof (*tp), 1);
    luaL_getmetatable (L, "VOLK.Term");
    lua_setmetatable (L, -2);

    return tp;
}


VOLK_TermSet **allocate_tset (lua_State *L)
{
    VOLK_TermSet **ts_p = lua_newuserdatauv (L, sizeof (*ts_p), 1);
    luaL_getmetatable (L, "VOLK.TermSet");
    lua_setmetatable (L, -2);

    return ts_p;
}


VOLK_rc
term_set_to_table (lua_State *L, VOLK_TermSet *ts)
{
    size_t i = 0;
    VOLK_Term *tmp = NULL;
    VOLK_rc rc;
    lua_newtable (L);
    while ((rc = VOLK_term_set_next (ts, &i, &tmp)) == VOLK_OK) {
        // Table keys are term hashes.
        lua_pushinteger (L, VOLK_term_hash (tmp));

        // Table values are the term sets proper.
        VOLK_Term **tp = (VOLK_Term **)lua_newuserdata (L, sizeof *tp);
        luaL_getmetatable (L, "VOLK.Term");
        lua_setmetatable (L, -2);
        *tp = VOLK_term_copy (tmp);

        lua_rawset (L, -3);
    }

    if (rc < VOLK_OK) return rc;
    return VOLK_OK;
}


/*
 * Factory methods.
 */

static int l_new_iriref (lua_State *L)
{
    const char *data = luaL_checkstring (L, 1);

    VOLK_Term **tp = allocate_term (L);

    *tp = VOLK_iriref_new (data);
    LUA_NLCHECK (*tp, "Error creating term.");

    return 1;
}


static int l_new_iriref_ns (lua_State *L)
{
    const char *data = luaL_checkstring (L, 1);

    VOLK_Term **tp = allocate_term (L);

    *tp = VOLK_iriref_new_ns (data);
    LUA_NLCHECK (*tp, "Error creating term.");

    return 1;
}


static int l_new_iriref_abs (lua_State *L)
{
    VOLK_Term
        *root = check_term (L),
        *iri = *(VOLK_Term **)luaL_checkudata (L, 2, "VOLK.Term");

    VOLK_Term **tp = allocate_term (L);

    *tp = VOLK_iriref_new_abs (root, iri);
    LUA_NLCHECK (*tp, "Error creating term.");

    return 1;
}


static int l_new_iriref_rel (lua_State *L)
{
    VOLK_Term
        *root = check_term (L),
        *iri = *(VOLK_Term **)luaL_checkudata (L, 2, "VOLK.Term");

    VOLK_Term **tp = allocate_term (L);

    *tp = VOLK_iriref_new_rel (root, iri);
    LUA_NLCHECK (*tp, "Error creating term.");

    return 1;
}


/** @brief create a new literal.
 *
 * Argument 2 (data type) and 3 (lang) are exclusive. If both are specified,
 * datatype has precedence. If both are nil, datatype is set to the default
 * (xsd:string).
 */
static int l_new_lit (lua_State *L)
{
    const char
        *data = luaL_checkstring (L, 1),
        *dtype_str = lua_tostring (L, 2);
    const char *lang = lua_tostring (L, 3);

    VOLK_Term **tp = allocate_term (L);
    VOLK_Term *dtype;

    if (dtype_str) {
        // TODO check memory leak.
        dtype = VOLK_iriref_new (dtype_str);
        *tp = VOLK_literal_new (data, dtype);
    }
    else if (lang) *tp = VOLK_lt_literal_new (data, (char *)lang);
    else *tp = VOLK_literal_new (data, NULL);

    LUA_NLCHECK (*tp, "Error creating term.");

    return 1;
}


static int l_new_bnode (lua_State *L)
{
    const char *data = lua_tostring (L, 1);

    VOLK_Term **tp = allocate_term (L);

    *tp = VOLK_bnode_new (data);
    LUA_NLCHECK (*tp, "Error creating term.");

    return 1;
}


static int l_new_copy (lua_State *L)
{
    VOLK_Term *src = check_term (L);

    VOLK_Term **tp = allocate_term (L);

    *tp = VOLK_term_copy (src);
    LUA_NLCHECK (*tp, "Error creating term.");

    return 1;
}


/*
 * Class methods.
 */
static int l_term_equals (lua_State *L)
{
    VOLK_Term
        *t1 = check_term (L),
        *t2 = *(VOLK_Term **)luaL_checkudata (L, 2, "VOLK.Term");

    lua_pushboolean (L, VOLK_term_equals (t1, t2));

    return 1;
}


static int l_term_gc (lua_State *L)
{
    VOLK_Term **tp = luaL_checkudata(L, 1, "VOLK.Term");
    //LOG_TRACE ("Garbage collecting term @%p", *tp);

    VOLK_term_free (*tp);
    *tp = NULL;

    return 0;
}


/*
static int l_term_to_string (lua_State *L)
{
    VOLK_Term *t = check_term (L);
    char *nt_term = NULL;

    CHECK (nt_codec.encode_term (t, NULL, &nt_term), fail);
    lua_pushfstring (L, "VOLK.Term @%p %s", t, nt_term);

    return 1;

fail:
    if (nt_term) free (nt_term);
    return luaL_error (L, "Error encoding term for display!");
}
*/


static int l_term_to_n3 (lua_State *L)
{
    VOLK_Term *t = check_term (L);
    char *nt_term = NULL;

    CHECK (nt_codec.encode_term (t, &nt_term), fail);
    lua_pushstring (L, nt_term);
    free (nt_term);

    return 1;

fail:
    if (nt_term) free (nt_term);
    return luaL_error (L, "Error encoding term for display.");
}


/*
 * Getters.
 */

// Forward declaration.
static const luaL_Reg term_getters [];


static int l_term_get_attr (lua_State *L)
{
    const char *attr = luaL_checkstring (L, 2);
    //printf ("Got attr: %s\n", attr);

    if (luaL_getmetafield (L, 1, attr) != LUA_TNIL)
        return 1;  // Return metamethod to be called as a function.

    luaL_getmetafield (L, 1, "getters");
    if (lua_getfield (L, -1, attr) != LUA_TNIL)
        return lua_tocfunction (L, -1)(L);

    return 0;
}


static int get_data (lua_State *L)
{
    VOLK_Term *t = check_term (L);
    lua_pushstring (L, t->data);

    return 1;
}


static int get_type (lua_State *L)
{
    VOLK_Term *t = check_term (L);
    lua_pushinteger (L, t->type);

    return 1;
}


static int get_iriref_prefix (lua_State *L)
{
    VOLK_Term *t = check_term (L);
    lua_pushstring (L, VOLK_iriref_prefix (t));

    return 1;
}


static int get_iriref_path (lua_State *L)
{
    VOLK_Term *t = check_term (L);
    lua_pushstring (L, VOLK_iriref_path (t));

    return 1;
}


static int get_iriref_frag (lua_State *L)
{
    VOLK_Term *t = check_term (L);
    lua_pushstring (L, VOLK_iriref_frag (t));

    return 1;
}


static int get_lit_datatype (lua_State *L)
{
    printf ("Getting datatype.\n");
    VOLK_Term *t = check_term (L);
    if (!VOLK_IS_LITERAL (t))
        return luaL_error (L, "Term is not a literal.", 2);

    lua_pushstring (L, t->datatype->data);

    return 1;
}


static int get_lit_lang (lua_State *L)
{
    VOLK_Term *t = check_term (L);
    if (!VOLK_IS_LITERAL (t))
        return luaL_error (L, "Term is not a literal.", 2);
    if (t->type != VOLK_TERM_LT_LITERAL) return 0;

    lua_pushstring (L, t->lang);

    return 1;
}


static int get_hash (lua_State *L)
{
    VOLK_Term *t = check_term (L);

    lua_pushinteger (L, VOLK_term_hash (t));

    return 1;
}


/*
 * Setters.
 */

// Very simple for now.
//static const luaL_Reg term_setters [];
static int l_term_set_attr (lua_State *L)
{ return luaL_error (L, "Direct setting is not allowed for this type.", 2); }


/*
 * Link map.
 */

static int lmap_iter_next (lua_State *L)
{
    VOLK_LinkMapIterator *it =
        *(VOLK_LinkMapIterator **)lua_touserdata (L, lua_upvalueindex (1));

    VOLK_Term **link_p = (VOLK_Term **)lua_newuserdata (L, sizeof *link_p);
    *link_p = NULL;
    luaL_getmetatable (L, "VOLK.Term");
    lua_setmetatable (L, -2);

    VOLK_TermSet *ts = NULL;
    VOLK_Term *tmp = NULL;
    VOLK_rc rc = VOLK_link_map_next (it, &tmp, &ts);

    if (rc != VOLK_OK) {
        if (rc == VOLK_END) return 0;
        else LUA_PCHECK (rc, "Error iterating over link map");
    }
    *link_p = VOLK_term_copy (tmp);
    LUA_NLCHECK (*link_p, "Error allocating term.");

    LUA_PCHECK (term_set_to_table (L, ts), "Error generating term set");

    // linked term + term set.
    return 2;
}


/**
 * Internally this function creates a LMapIterator, which is used as the
 * upvalue for lmap_iter_next(). The iterator is garbage collected at the end
 * of the iteration loop, the link map can be reused.
 */
static int l_lmap_iter_init (lua_State *L)
{
    stack_dump (L, "beginning of LMap iter init fn");
    VOLK_LinkMap *lm = *(VOLK_LinkMap **)luaL_checkudata(L, 1, "VOLK.LinkMap");

    VOLK_LinkMapIterator **lmit_p =
        (VOLK_LinkMapIterator  **)lua_newuserdata (L, sizeof *lmit_p);
    *lmit_p = VOLK_link_map_iter_new (lm);
    LUA_NLCHECK (*lmit_p, "Error creating Link map iterator.");
    luaL_getmetatable (L, "VOLK.LMapIterator");
    lua_setmetatable (L, -2);

    lua_pushcclosure (L, lmap_iter_next, 1);
    stack_dump (L, "After pushing iter closure");

    return 1;
}


static int link_map_gc (lua_State *L)
{
    VOLK_LinkMap **lm_p = lua_touserdata (L, 1);
    LOG_DEBUG ("Garbage collecting link map @%p", *lm_p);
    // FIXME This is to prevent a double-free on shutdown.
    // Must find the culprit instead.
    if (UNLIKELY (!lm_p || !*lm_p)) return 0;

    VOLK_link_map_free (*lm_p);
    *lm_p = NULL;

    return 0;
}


static int lmap_iter_gc (lua_State *L)
{
    VOLK_LinkMapIterator *it = *(VOLK_LinkMapIterator **)lua_touserdata (L, 1);
    LOG_DEBUG ("Garbage collecting link map iterator @%p", it);
    VOLK_link_map_iter_free (it);

    return 0;
}


/*
 * Library setup.
 */

static const luaL_Reg term_lib_fn [] = {
    {"new_iriref", l_new_iriref},
    {"new_iriref_ns", l_new_iriref_ns},
    {"new_iriref_abs", l_new_iriref_abs},
    {"new_iriref_rel", l_new_iriref_rel},
    {"new_lit", l_new_lit},
    {"new_bnode", l_new_bnode},
    {"new_copy", l_new_copy},

    {NULL}
};


static const luaL_Reg term_getters [] = {
    // General getters.
    {"data", get_data},
    {"type", get_type},

    // IRIRef getters.
    {"prefix", get_iriref_prefix},
    {"path", get_iriref_path},
    {"frag", get_iriref_frag},

    // Literal getters.
    {"datatype", get_lit_datatype},
    {"lang", get_lit_lang},

    // Generic getters.
    {"hash", get_hash},

    {NULL}
};


/*
static const luaL_Reg term_setters [] = {
    {NULL}
};
*/

static const luaL_Reg term_lib_meth [] = {
    {"__eq", l_term_equals},
    {"__gc", l_term_gc},
    {"__index", l_term_get_attr},
    {"__tostring", l_term_to_n3},
    {"__newindex", l_term_set_attr},

    {NULL}
};


static const LEnumConst term_enums[] = {
    {"TYPE_IRIREF", VOLK_TERM_IRIREF},
    {"TYPE_LITERAL", VOLK_TERM_LITERAL},
    {"TYPE_LT_LITERAL", VOLK_TERM_LT_LITERAL},
    {"TYPE_BNODE", VOLK_TERM_BNODE},

    {"LINK_INBOUND", VOLK_LINK_INBOUND},
    {"LINK_OUTBOUND", VOLK_LINK_OUTBOUND},
    {"LINK_EDGE", VOLK_LINK_EDGE},

    {NULL, 0}
};


static const LStringConst term_strings[] = {
    {"RDF_TYPE", VOLK_RDF_TYPE},
    {"RDF_TYPE_NS", VOLK_RDF_TYPE_NS},
    {"DEFAULT_DTYPE", DEFAULT_DTYPE},
    {"DEFAULT_DTYPE_NS", DEFAULT_DTYPE_NS},
    {"DEFAULT_CTX", DEFAULT_CTX_LABEL},

    {NULL, 0}
};


int luaopen_volksdata_term (lua_State *L)
{
    VOLK_init();  // This is idempotent: no problem calling it multiple times.
    luaL_newmetatable (L, "VOLK.Term");
    luaL_setfuncs (L, term_lib_meth, 0);

    // Getters table.
    lua_newtable (L);
    for (int i = 0; term_getters[i].name != NULL; i++) {
        lua_pushcfunction (L, term_getters[i].func);
        lua_setfield (L, -2, term_getters[i].name);
    }
    // Set getters table as a value for the Term metatable.
    lua_setfield (L, -2, "getters");

    /*
     * Metatables for ancillary types.
     */
    // Link map.
    luaL_newmetatable (L, "VOLK.LinkMap");
    lua_pushcfunction (L, link_map_gc);
    lua_setfield (L, -2, "__gc");
    lua_pushcfunction (L, l_lmap_iter_init);
    lua_setfield (L, -2, "__pairs");

    // Link map iterator.
    luaL_newmetatable (L, "VOLK.LMapIterator");
    lua_pushcfunction (L, lmap_iter_gc);
    lua_setfield (L, -2, "__gc");

    luaL_newlib (L, term_lib_fn);

    // Module-level constants.
    push_string_const (L, term_strings);
    push_int_const (L, term_enums);

    return 1;
}
