#include "lua_volksdata.h"


/*
 * Factory methods.
 */

static int l_triple_new (lua_State *L)
{
    VOLK_Triple **trp_p = lua_newuserdatauv (L, sizeof (*trp_p), 1);
    luaL_getmetatable (L, "VOLK.Triple");
    lua_setmetatable (L, -2);
    VOLK_Term
        *s = *(VOLK_Term **)luaL_checkudata(L, 1, "VOLK.Term"),
        *p = *(VOLK_Term **)luaL_checkudata(L, 2, "VOLK.Term"),
        *o = *(VOLK_Term **)luaL_checkudata(L, 3, "VOLK.Term");

    *trp_p = VOLK_triple_new (
            VOLK_term_copy (s),
            VOLK_term_copy (p),
            VOLK_term_copy (o));
    if (!*trp_p) return luaL_error (L, "Error while creating a triple!");

    return 1;
}


/*
 * Class methods.
 */

static int l_triple_gc (lua_State *L)
{
    VOLK_Triple *trp = check_triple (L);
    if (trp) VOLK_triple_free (trp);

    return 0;
}


/** @brief Get information about a triple's term.
 *
 * Note that this is only a visual representation, as terms cannot be modified
 * outside of their containing triple (unlike in the C library).
 *
 * @FIXME this doesn't allow chaining access methods, e.g. trp.s.data
 *
 * To get an independent copy of a triple term, use triple.copy_term().
 */
static int l_triple_get_term (lua_State *L)
{
    const VOLK_Triple *trp = check_triple (L);
    const char pos = luaL_checkstring (L, 2)[0];
    VOLK_TriplePos pn;
    if (pos == 's') pn = TRP_POS_S;
    else if (pos == 'p') pn = TRP_POS_P;
    else if (pos == 'o') pn = TRP_POS_O;
    else return luaL_error (L, "Out of range position: %c.", pos);

    VOLK_Term **tp = allocate_term (L);
    *tp = VOLK_term_copy (VOLK_triple_pos (trp, pn));
    if (!*tp) return luaL_error (L, "Error while copying a triple term!");

    return 1;
}


static int l_triple_set_term (lua_State *L)
{
    VOLK_Triple *trp = check_triple (L);
    const char pos = luaL_checkstring (L, 2)[0];
    const VOLK_Term *t = *(VOLK_Term **)luaL_checkudata(L, 3, "VOLK.Term");

    VOLK_Term *new_t = VOLK_term_copy (t);

    if (pos == 's') {
        VOLK_term_free (trp->s);
        trp->s = new_t;
    } else if (pos == 'p') {
        VOLK_term_free (trp->p);
        trp->p = new_t;
    } else if (pos == 'o') {
        VOLK_term_free (trp->o);
        trp->o = new_t;
    } else return luaL_error(L, "Out of range position: %c.", pos);

    return 1;
}


static int l_triple_copy_term (lua_State *L)
{
    const VOLK_Triple *trp = check_triple (L);
    const char pos = luaL_checkstring (L, 2)[0];
    VOLK_TriplePos pn;
    if (pos == 's') pn = TRP_POS_S;
    else if (pos == 'p') pn = TRP_POS_P;
    else if (pos == 'o') pn = TRP_POS_O;
    else return luaL_error(L, "Copy: Out of range position: %c.", pos);

    VOLK_Term **tp = allocate_term (L);
    *tp = VOLK_term_copy (VOLK_triple_pos (trp, pn));
    if (!*tp) return luaL_error (L, "Error while copying a triple term!");

    return 1;
}


static int l_triple_to_n3 (lua_State *L)
{
    const VOLK_Triple *trp = check_triple (L);
    char *nt_term = NULL;

    for (int i = TRP_POS_S; i <= TRP_POS_O; i++) {
        CHECK (nt_codec.encode_term (
                    VOLK_triple_pos (trp, i), &nt_term), fail);
        lua_pushstring (L, nt_term);
        lua_pushstring (L, " ");
    }
    free (nt_term);
    lua_pushstring (L, ".");  // N3 statement termination.
    lua_concat (L, 7);

    return 1;

fail:
    if (nt_term) free (nt_term);
    return luaL_error (L, "Error encoding triple term.");
}


/*
 * Library setup.
 */

static const struct luaL_Reg triple_lib_fn [] = {
    {"new", l_triple_new},
    {NULL}
};


static const struct luaL_Reg triple_lib_meth [] = {
    {"__gc", l_triple_gc},
    {"__index", l_triple_get_term},
    {"__newindex", l_triple_set_term},
    {"__tostring", l_triple_to_n3},

    {"copy_term", l_triple_copy_term},

    {NULL}
};


static const LEnumConst triple_enums[] = {
    {"POS_S", TRP_POS_S},
    {"POS_P", TRP_POS_P},
    {"POS_O", TRP_POS_O},
    {NULL, 0}
};


int luaopen_volksdata_triple (lua_State *L)
{
    VOLK_init();  // This is idempotent: no problem calling it multiple times.
    luaL_newmetatable (L, "VOLK.Triple");
    lua_pushvalue (L, -1);

    // MT
    lua_setfield (L, -1, "__index");
    luaL_setfuncs (L, triple_lib_meth, 0);

    luaL_newlib (L, triple_lib_fn);

    // Module-level constants.
    push_int_const (L, triple_enums);

    return 1;
}

